@make(ais)
@aiInfo(topic=<[Replaced by AI-00546]>,
number=[AI-00048/13],
classcode=[na],
statuscode=[na],
mainref=[14.03.02@\(01)],
date=[93-06-24])
@aiHeading()
@begin(history)
@begin(toplines)
@mainref()
@standard[14.03.01@\(05)]
!class @cbon()no action 93-06-18 (was: BI 86-02-21)
!status no action 93-06-18 (see AI-00546)@cboff()
@end(toplines)
!status approved by WG9/AJPO 87-02-20
!status approved by Director, AJPO 87-02-20
!status approved by Ada Board (18-0-3) 87-02-19
!status approved by WG9 86-05-09
!status committee-approved 86-02-21 (8-0-0)
!status reconsidered by committee 86-02-20 (7-0-1)
!status committee-approved (8-2-0) 85-11-20 (pending editorial review)
!status work-item 85-03-04
!status returned to committee by WG9/ADA Board 85-02-26
!status committee-approved 84-11-28
!status work-item 84-06-29
!status board returned to committee 84-06-29
!status committee-approved 84-02-06
!status work-item 84-01-29
@cbon()!references AI-00546, 83-00033
@topic{!topic [Replaced by AI-00546]}@cboff()
@end(history)

@subheading[!summary 86-03-05]

The CLOSE operation can be applied to a file object that is also serving as
the default input or default output file.  The effect is to close the default
file.  A subsequent OPEN operation can have the effect of opening the default
file as well.  Similarly, a DELETE operation can be applied to a file object
that is serving as the default file.

The exception MODE_@|ERROR is raised by OPEN if the specified mode is OUT_@|FILE
and the file object being opened is serving as the default input file.
Similarly, MODE_@|ERROR is raised if the specified mode is IN_@|FILE and the file
object being opened is serving as the default output file.

@subheading[!question 85-01-02]

Although it is not possible to use RESET to change the mode of a file that is
serving as a default input or output file, is it possible to close the file
and reopen it with a different mode?   Is it possible to delete the current
default file?

@subheading[!recommendation 86-04-02]

The CLOSE operation can be applied to a file object that is also serving as
the default input or default output file.  The effect is to close the default
file.  A subsequent OPEN operation can have the effect of opening the default
file as well.  Similarly, a DELETE operation can be applied to a file object
that is serving as the default file.

The exception MODE_@|ERROR is raised by OPEN if the specified mode is OUT_@|FILE
and the file object being opened is serving as the default input file.
Similarly, MODE_@|ERROR is raised if the specified mode is IN_@|FILE and the file
object being opened is serving as the default output file.

@subheading[!discussion 86-04-02]

Consider the following:
@begin(example)
@key[declare]
    F : TEXT_@|IO.FILE_@|TYPE;
@key[begin]
    OPEN (F, IN_@|FILE, "FOO");
    SET_@|INPUT (F);
    CLOSE (F);                    -- is the default file closed?
@hinge()

    OPEN (F, OUT_@|FILE, "BAR");    -- what is the default file now?
    GET (CURRENT_@|INPUT, ...);     -- is there an exception?
@key[end];
@end(example)
Is the close operation on F permitted even though F is serving as the default
input file?  If so, is the default file also closed, or is it still
associated with external file @dq"FOO"?  If the open operation is successful, is
the default input file associated with external file @dq"BAR" with mode
OUT_@|FILE?  Is the default file associated with any external file after
the OPEN operation?

To answer these questions, we need to consider several definitions given in
the Standard.  14.1(2) says:
@begin(quotation)
Input and output operations are expressed as operations on objects of
some @i[file type], rather than directly in terms of the external
files.  In the remainder of this chapter, the term @i[file] is always
used to refer to a file object; the term @i[external file] is used
otherwise.
@end(quotation)
To make the discussion clearer here, when quoting from the Standard, we will
substitute the term @dq"file object" when the Standard uses just the term
@dq"file."

14.1(6) goes on to describe the conceptual model underlying the I/O
operations:
@begin(quotation)
Before input or output operations  can be performed on a  [file
object], the [file  object] must  first be  associated with  an
external file.  While  such an  association is  in effect,  the
[file object] is  said to  be  OPEN, and  otherwise  the  [file
object] is said to be CLOSED.
@end(quotation)
In other words, the effect of executing an open operation is just to
establish an association between a file object and an external file.  (Of
course, while this association is in effect, operations performed on the file
object will affect (and be affected by) the external file.)

The semantics of the open operation for text files are given in 14.3.1(1-2)
and 14.2.1(6), which says:
@begin(quotation)
[OPEN] associates  the given  [file  object] with  an  existing
external file  having the  given name  and form,  and sets  the
current mode of the given [file object] to the given mode.  The
given [file object] is left open.
@end(quotation)
The definition of the close operation is given in 14.3.1(1, 3) and 14.2.1(9),
which says:
@begin(quotation)
[CLOSE] severs the association between the given [file  object]
and its associated external file.   The given [file object]  is
left closed.
@end(quotation)
Finally, the definition for SET_@|INPUT says [14.3.2(3)]:
@begin(quotation)
[SET_@|INPUT] sets  the current  default input  [file object]  to
FILE.
@end(quotation)
The effect of SET_@|INPUT (and similarly, of course, SET_@|OUTPUT) is to make the
default input file object identical with the file object that is the value of
SET_@|INPUT's argument, i.e., the default input file is not independently
associated with any external file, but rather shares the file object that is
specified by SET_@|INPUT.  After executing SET_@|INPUT(F), CURRENT_@|INPUT and F
have the same file object value.  This means operations on F and operations on
the default input file have exactly the same effect, e.g., NEW_@|LINE(F) changes
the line count for the default input file.

When the operation CLOSE(F) is executed, the association between the shared
file object and the external file is severed, but there is no implication that
the file object serving as the default input file somehow loses its
association with the file object that is the value of F.  Instead, the effect
is simply to sever the external file's association with the file object
associated with both F and CURRENT_@|INPUT.  A similar effect occurs for DELETE.
DELETE(F) deletes the associated external file and also severs the association
between the file object named by F (which is also the default file object)
and the external file.

A subsequent open operation that uses F's file object establishes an
association between this file object and an external file.  Since this file
object is still serving as the default input file object, the association is
established for CURRENT_@|INPUT as well.  If the OPEN operation in the example
had specified mode IN_@|FILE, then the subsequent GET operation would apply to
the external file @dq"BAR".  The effect would be the same as if file object F
had been named in the argument instead of CURRENT_@|INPUT.

Of course, the OPEN operation in the example attempts to establish an
OUT_@|FILE association between F's file object and the external file.  This
poses a problem since F's file object is also the current default input file
object and the Standard, in general, tries to ensure that the mode of a
default file object is consistent with its use as an input or output file.
In particular, SET_@|INPUT and SET_@|OUTPUT raise MODE_@|ERROR if the mode of the
file object being established as a default file is not consistent with the
file's use for input or output [14.3.2(4, 7)].  Similarly, RESET raises
MODE_@|ERROR upon an attempt to change the mode of a file that is either the
current default input file object or the current default output file object
[14.3.1(5)].  To ensure a default file object has an appropriate mode,
MODE_@|ERROR should also be raised by OPEN if the file object being associated
with an external file is serving as a default file object and the mode
specified by OPEN is inconsistent with the file object's use as a default
input or output file.  (So in the example, MODE_@|ERROR should be raised by
OPEN since the file object denoted by F is serving as the default input file
object.)

An objection to allowing a default file object to be closed and subsequently
reopened is that it forces a somewhat more complicated implementation.
Suppose an implementation wants to represent a file object as a record.  It
can't use a simple record representation because SET_@|INPUT and SET_@|OUTPUT
require that file objects be shared.  So the implementation must use a
pointer to a record.  Given that pointers must be used, it would be
convenient to represent a closed file with a null pointer, e.g., when F's
declaration is elaborated, its real value is set to null.  The open operation
allocates and initializes a record object (as well as gaining access to the
external file).  The close operation releases access to the external file and
sets the pointer to null (e.g., CLOSE (F) would set F's value to null).  Such
an implementation could release the storage used by the designated record
object if this object were not designated by any other pointer, which would
be the case if F's file object was not also being used as a default file
object.  When F's file object is also a default file object, the close
operation cannot assign null to F because a subsequent open operation for F
must also establish an external file association for the default file object,
i.e., the designated record object used for F's new external file association
must still be the same as the default file object.

In short, a somewhat more efficient implementation would be possible if
closing a default file object implied the default file object was no longer
shared with the value of any file variable.  But the Standard does not allow
such an interpretation (and the additional implementation burden is not
sufficiently great to warrant considering such an interpretation further.)

